from numbers import Integral
from pydoc import describe

from mathutils import Color
import bpy

from . properties import *
from . handlers import *


class PA_PlanetData(PropertyGroup):
    ############################################################################
    # General planet-specific data
    ############################################################################

    name: StringProperty(
        name="",
        description="Name of the given planet data",
        default="Planet",
    )

    node_tree: PointerProperty(
        name="Node Tree",
        description="Node tree for planet data",
        type=bpy.types.NodeTree,
    )


    ### Atmosphere ---------------------------------

    atmosphere_type: EnumProperty(
        name="Atmosphere",
        description="Enable atmosphere",
        items=[
            ("ENABLED", "Enabled",
                "Atmosphere enabled"),
            ("DISABLED", "Disabled",
                "No atmosphere enabled"),
        ],
        default='DISABLED',
        update=update_atmosphere_type,
    )

    atmosphere_radius: FloatProperty(
        name="Atmosphere Radius",
        description="Radius of atmosphere from planet's surface to it's outer bounds",
        unit='LENGTH',
        min=0.0,
        default=1.0,
    )

    atmosphere_density: FloatProperty(
        name="Atmosphere Density",
        description="Density of atmosphere",
        min=0.0,
        default=1.2,
    )

    atmosphere_rayleigh_scale_height: FloatProperty(
        name="Rayleigh Scale Height",
        description="Determines shape of Rayleigh scattering density curve",
        subtype='FACTOR',
        min=0.0,
        soft_max=1.0,
        default=0.08,
    )

    atmosphere_rayleigh_coefficient: FloatVectorProperty(
        name="Rayleigh Inscattering",
        description="Atmosphere scattering coefficients for Rayleigh scattering",
        subtype='COLOR',
        default=Color((0.058, 0.136, 0.331)),
        soft_min=0,
        soft_max=1,
        size=3,
    )

    atmosphere_mie_scale_height: FloatProperty(
        name="Mie Scale Height",
        description="Determines shape of Mie scattering density curve",
        subtype='FACTOR',
        min=0.0,
        soft_max=1.0,
        default=0.07,
    )

    atmosphere_mie_coefficient: FloatVectorProperty(
        name="Mie Inscattering",
        description="Atmosphere scattering coefficients for Mie scattering",
        subtype='COLOR',
        default=Color((0.04, 0.04, 0.04)),
        soft_min=0,
        soft_max=1,
        size=3,
    )

    atmosphere_noise_jitter: FloatProperty(
        name="Atmosphere Noise Jitter",
        description="Amount of noise jitter used to increase planet accuracy. Adds noise",
        subtype='FACTOR',
        min=0.0,
        soft_max=0.2,
        default=0.1,
    )

    atmosphere_reflection_jitter_multiplier: FloatProperty(
        name="Atmosphere Reflection Jitter Multiplier",
        description="Scale noise in reflections, helpful for Eevee where reflections don't get sampled continuously",
        subtype='FACTOR',
        min=0.0,
        soft_max = 0.2,
        default=0.02,
    )

    ### --------------------------------------------

    ### Rings --------------------------------------

    rings_type: EnumProperty(
        name="Rings Type",
        description="Type of the rings",
        items=[
            ("PROCEDURAL", "Procedural",
                "Use procedural noise as the ring texture"),
            ("TEXTURE", "Texture",
                "Use an image texture (color + alpha) as the ring texture"),
            ("DISABLED", "Disabled",
                "Disable rings")
        ],
        default='DISABLED',
        update=update_ring_type,
    )

    rings_vector: FloatVectorProperty(
        name="Rotation",
        description="Ring rotation vector",
        subtype='DIRECTION',
        default=Vector((0.0, 0.0, 1.0))
    )

    rings_start: FloatProperty(
        name="Innder Radius",
        description="Radius of rings closest to the planet, relative to planet radius",
        min=0.0,
        default=1.5,
    )

    rings_end: FloatProperty(
        name="Outer Radius",
        description="Radius of rings furthest from the planet, relative to planet radius",
        min=0.0,
        default=3.0,
    )

    rings_indirect_shadow: FloatProperty(
        name="Indirect Shadow Factor",
        description="Controls the shape of the shadow effect on indirect light",
        subtype='FACTOR',
        min=0.0,
        soft_max=4.0,
        default=2.0,
    )

    rings_translucency: FloatProperty(
        name="Translucency",
        description="Translucency of ring material",
        subtype='FACTOR',
        min=0.0,
        max=1.0,
        default=0.1,
    )

    rings_tint: FloatVectorProperty(
        name="Color Tint",
        description="Color Tint of rings",
        subtype='COLOR',
        default=Color((0.5, 0.5, 0.5)),
        min=0.0,
        max=1.0,
        size=3,
    )

    rings_texture: PointerProperty(
        name="Rings Texture",
        description="Ring texture map, using alpha channel",
        type=bpy.types.Image,
        update=update_ring_texture,
    )

    ring_procedural_texture_scale: FloatProperty(
        name="Scale",
        description="Scale of the procedural noise texture",
        min=0.0,
        default=15.0,
    )

    ring_procedural_texture_detail: FloatProperty(
        name="Detail",
        description="Detail amount of the procedural noise texture",
        min=0.0,
        default=3.0,
    )

    ring_procedural_texture_roughness: FloatProperty(
        name="Roughness",
        description="Roughness of the procedural noise texture",
        subtype='FACTOR',
        min=0.0,
        soft_max=1.0,
        default=0.5,
    )

    ### --------------------------------------------


    ### Materiality --------------------------------

    # General values -------------------------------

    indirect_color: FloatVectorProperty(
        name="Indirect Color",
        description="Indirect color of planet, used for bounce lighting calculations",
        subtype='COLOR',
        default=Color((0.5, 0.5, 0.5)),
        min=0.0,
        max=1.0,
        size=3,
    )

    scale: FloatProperty(
        name="Scale",
        description="Scale of procedural surface textures",
        min=0.000001,
        default=5.0,
    )

    detail_scale: FloatProperty(
        name="Detail Scale",
        description="Multiplier for scale of procedural surface texture details",
        min=0.000001,
        default=3.0,
    )

    terrain_height: FloatProperty(
        name="Terrain Height",
        description="Average height offset of the terrain relief",
        min=0.0,
        default=2.0,
    )

    #-----------------------------------------------


    # Textured values ------------------------------
    albedo: FloatVectorProperty(
        name="Albedo",
        description="Planet surface albedo",
        subtype='COLOR',
        default=Color((0.8, 0.8, 0.8)),
        min=0.0,
        max=1.0,
        size=3,
    )

    albedo_texture: PointerProperty(
        name="Albedo Texture",
        description="Planet surface albedo texture",
        type=bpy.types.Image,
        update=update_albedo_texture,
    )

    roughness: FloatProperty(
        name="Roughness",
        description="Planet surface specular roughness",
        subtype='FACTOR',
        default=0.8,
        min=0.0,
        max=1.0,
    )

    roughness_texture: PointerProperty(
        name="Roughness Texture",
        description="Planet surface specular roughness texture",
        type=bpy.types.Image,
        update=update_roughness_texture,
    )

    height_scale: FloatProperty(
        name="Height Scale",
        description="Scaling of height map influence",
        subtype='FACTOR',
        default=1.0,
        soft_min=0.0,
        soft_max=2.0,
    )

    height_texture: PointerProperty(
        name="Height Texture",
        description="Planet surface height texture",
        type=bpy.types.Image,
        update=update_height_texture,
    )

    emission: FloatVectorProperty(
        name="Emission",
        description="Planet surface emission",
        subtype='COLOR',
        default=Color((0.0, 0.0, 0.0)),
        min=0.0,
        max=1.0,
        size=3,
    )

    emission_texture: PointerProperty(
        name="Emission Texture",
        description="Planet surface emission texture",
        type=bpy.types.Image,
        update=update_emission_texture,
    )

    emission_strength: FloatProperty(
        name="Emission Strength",
        description="Intensity of planet surface emission",
        subtype='FACTOR',
        default=1.0,
        min=0.0,
        soft_max=10.0,
    )
    #-----------------------------------------------

    ###---------------------------------------------


class PA_PlanetInstance(PropertyGroup):
    ############################################################################
    # An instance of a planet
    ############################################################################


    name: StringProperty(
        name="",
        description="Name of the given planet instance",
        default="Planet",
    )

    ### Data ---------------------------------------

    planet_data: PointerProperty(
        name="Planet Data",
        description="Planet surface data",
        type=PA_PlanetData,
    )

    ### --------------------------------------------

    ### Transforms ---------------------------------
    transform_mode: EnumProperty(
        name="Transform Mode",
        description="Transform mode used by current planet instance",
        items=[
            ("GLOBAL", "Global",
                "Global transform based positioning, uses position, rotation and radius"),
            ("VIEWER", "Viewer-based",
                "Viewer-based transform, uses azimuth, elevation, distance, rotation and radius"),
            ("ORBIT", "Elliptical orbit",
                "Orbital positioning, uses parent transform, orbit plane, orbit ellipse values, orbit's rotation, planet's location on orbit, rotation and radius"),
        ],
        update=update_transform_mode,
    )

    # Shared transforms ----------------------------
    radius: FloatProperty(
        name="Radius",
        description="Planet's radius",
        unit='LENGTH',
        min=0.0001,
        soft_min=0.01,
        default=5.0,
    )

    rotation: FloatVectorProperty(
        name="Rotation",
        description="The rotation of given planet",
        subtype='EULER',
        size=3,
    )
    #-----------------------------------------------

    # GLOBAL transform -----------------------------
    position: FloatVectorProperty(
        name="Position",
        description="The position of given planet",
        subtype='TRANSLATION',
        size=3,
    )
    #-----------------------------------------------

    # VIEWER transfrom------------------------------
    azimuth: FloatProperty(
        name="Azimuth",
        description="Planet's azimuthal direction from viewer",
        subtype='ANGLE',
    )

    elevation: FloatProperty(
        name="Elevation",
        description="Planet's elevational direction from viewer",
        subtype='ANGLE',
    )

    distance: FloatProperty(
        name="Distance",
        description="Planet's distance from viewer",
        subtype='DISTANCE',
        default=100.0,
    )
    #-----------------------------------------------

    # ORBIT transfrom------------------------------
    ellipse_width: FloatProperty(
        name="Ellipse Width",
        description="Orbit ellipse width",
        unit='LENGTH',
        min=0.0001,
        soft_min=0.01,
        default=1.0,
    )

    ellipse_height: FloatProperty(
        name="Ellipse Height",
        description="Orbit ellipse height",
        unit='LENGTH',
        min=0.0001,
        soft_min=0.01,
        default=1.0,
    )

    orbit_rotation: FloatVectorProperty(
        name="Orbit Rotation",
        description="Orbit's rotation relative to its parent",
        subtype='EULER',
        size=3,
    )

    orbit_point_angle: FloatProperty(
        name="Orbit Point Angle",
        description="Planet's location on ellipse as angle",
        subtype='ANGLE'
    )

    orbit_parent_type: EnumProperty(
        name="Orbit Parent Type",
        description="The type of orbit origin location parent",
        items=[
            ("PLANET", "Planet", "A Planet instance"),
            ("OBJECT", "Object", "An object"),
            ("POINT", "Point", "A point vector"),
        ],
        update=update_orbit_origin,
    )

    orbit_parent_object: PointerProperty(
        name="Orbit Parent Object",
        description="Object used for determining origin point of orbit",
        type=bpy.types.Object,
        update=update_orbit_origin,
    )

    orbit_parent_planet: IntProperty(
        name="Orbit Parent Planet",
        description="Index of planet instance used for determining origin point of orbit",
        default=0,
        update=update_orbit_origin,
    )

    orbit_parent_point: FloatVectorProperty(
        name="Orbit Parent Point",
        description="Global vector point used as the origin point of orbit",
        subtype='TRANSLATION',
        size=3,
        update=update_orbit_origin,
    )

    #-----------------------------------------------

    ###---------------------------------------------
