import bpy
from bpy.app.handlers import persistent

from . helpers import *
from . planets import *


# Names for node groups used in addon
PHYSICAL_PLANETS_NAME = "PCO_PhysicalPlanets"
PHYSICAL_INPUTS_NAME = "PCO_PhysicalInputs"


def enable_celestial_objects(self, context):
    pco_gsettings = context.scene.world.pco_general_settings
    node_tree = context.scene.world.node_tree

    if pco_gsettings.enabled:
        active_output = node_tree.get_output_node('ALL')
        celestial_group = load_and_create_node_group(PHYSICAL_PLANETS_NAME, PHYSICAL_PLANETS_NAME)
        context.scene.world.pco_exposed.celestial_node = celestial_group.node_tree
        context.scene.world.pco_exposed.celestial_node_name = celestial_group.name
        pco_gsettings.use_standalone_lighting = not check_psa_present(context)
        link_celestial_object_group(context, celestial_group, active_output)
        # Prevent crashes on re-enable. Properly initialize existing data
        update_node_tree(context)
    else:
        remove_pco_node_groups(context)
        purge_node_groups()


def check_psa_present(context):
    return hasattr(context.scene.world, 'psa_general_settings') and context.scene.world.psa_general_settings.enabled


def link_celestial_object_group(context, celestial_group, active_output):
    node_tree = context.scene.world.node_tree
    
    if celestial_group is not None:
        if not context.scene.world.pco_general_settings.use_standalone_lighting:
            # Physical Starlight and Atmosphere is enabled
            psa_node = node_tree.nodes[context.scene.world.psa_exposed.atmosphere_node_name]
            stars_node = get_input_node(psa_node, 4)
            if stars_node is not None:
                node_tree.links.new(stars_node.outputs[0], celestial_group.inputs[0])
                node_tree.links.new(celestial_group.outputs[0], psa_node.inputs[4])
            else:
                node_tree.links.new(celestial_group.outputs[0], psa_node.inputs[4])
            move_node_in_between(stars_node, celestial_group, psa_node)
        else:
            pre_output = get_previous_node(active_output)
            if pre_output is not None:
                if pre_output.type in ('BACKGROUND', 'EMISSION'):
                    pre_pre_output = get_previous_node(pre_output)
                    put_node_on_link(node_tree, celestial_group, pre_output, active_output)
                    if pre_pre_output is not None:
                        link_nodes(node_tree, pre_pre_output, celestial_group)
                else:
                    pre_pre_output = get_previous_node(pre_output)
                    put_node_on_link(node_tree, pre_output, celestial_group, active_output)
                    if pre_pre_output is not None:
                        link_nodes(node_tree, pre_pre_output, pre_output)
            else:
                link_nodes(node_tree, celestial_group, active_output)
        pco_inputs = load_and_create_node_group(PHYSICAL_INPUTS_NAME, PHYSICAL_INPUTS_NAME)
        context.scene.world.pco_exposed.inputs_node = pco_inputs.node_tree
        if pco_inputs is not None:
            move_node_in_between(None, pco_inputs, celestial_group)
            node_tree.links.new(pco_inputs.outputs[0], celestial_group.inputs[1])
            node_tree.links.new(pco_inputs.outputs[1], celestial_group.inputs[2])
            node_tree.links.new(pco_inputs.outputs[2], celestial_group.inputs[3])


def remove_pco_node_groups(context):
    delete_node_group(PHYSICAL_PLANETS_NAME)
    delete_node_group(PHYSICAL_INPUTS_NAME)


def update_sun(self, context):
    init_sun_data(context)


def set_standalone_lighting(self, context):
    pass


def update_node_tree(context):
    # Cleanup the tree
    cleanup_celestial_node_tree(context)

    # Add all nodes
    add_planet_transform_nodes(context)
    add_planet_data_nodes(context)
    add_planet_depth_sort_nodes(context)
    add_planet_shadow_nodes(context)
    add_planet_indirect_nodes(context)
    add_ring_nodes(context)
    add_atmosphere_nodes(context)

    # Link all nodes
    link_planet_data_nodes(context)
    link_depth_sort_nodes(context)
    link_indirect_light_nodes(context)
    link_shadow_nodes(context)
    link_ring_node(context)
    link_atmosphere_node(context)
    link_planet_shader_node(context)
    link_celestial_output_node(context)

    # Organize node tree
    organize_nodes(context)

    # Enable all drivers
    reinitialize_drivers(context)

    # Set all textures
    set_all_textures(context)


def reinitialize_drivers(context):
    add_transform_drivers(context)
    add_textured_material_drivers(context)
    add_atmosphere_drivers(context)
    add_ring_drivers(context)


def update_atmosphere_type(self, context):
    world = context.scene.world
    if world.pco_planet_slots[world.pco_active_planet_index].planet_data.atmosphere_type != 'DISABLED':
        world.pco_exposed.atmosphere_instance = world.pco_active_planet_index
    else:
        world.pco_exposed.atmosphere_instance = -1
    update_node_tree(context)


def update_ring_type(self, context):
    world = context.scene.world
    if world.pco_planet_slots[world.pco_active_planet_index].planet_data.rings_type != 'DISABLED':
        world.pco_exposed.ring_instance = world.pco_active_planet_index
    else:
        world.pco_exposed.ring_instance = -1
    update_node_tree(context)
    update_ring_texture(self, context)


def update_albedo_texture(self, context):
    set_albedo_texture(context, context.scene.world.pco_active_planet_index)


def update_roughness_texture(self, context):
    set_roughness_texture(context, context.scene.world.pco_active_planet_index)


def update_height_texture(self, context):
    set_height_texture(context, context.scene.world.pco_active_planet_index)


def update_emission_texture(self, context):
    set_emission_texture(context, context.scene.world.pco_active_planet_index)


def update_ring_texture(self, context):
    set_ring_texture(context, context.scene.world.pco_active_planet_index)


def update_transform_mode(self, context):
    add_transform_drivers(context)
    link_planet_data_nodes(context)


def update_orbit_origin(self, context):
    add_transform_drivers(context)
    link_planet_data_nodes(context)


def purge_node_groups():
    """ Delete all Node Groups from the blenderfile """

    # List of shader node group prefixes in the addon
    group_name_prefixes = ('PCO_', '. PCO_')

    for g in bpy.data.node_groups:
        if any(g.name.startswith(s) for s in group_name_prefixes):
            bpy.data.node_groups.remove(g)
