import bpy
import os

from mathutils import Vector

# .blend with node groups
PCO_BLEND_FILE = "celestial_objects.blend"


def link_driver_simple(source, target, property, data_path, index=-1, id_type='WORLD'):
    # Adds simple driver linking the property on source with target's dataPath
    # Use index > 0 if you need to indicate the target channel of property that the driver affects
    # id_type is the type of target
    d = source.driver_add(property, index).driver

    for var in d.variables:
        d.variables.remove(var)

    v = d.variables.new()
    v.name = property
    v.targets[0].id_type = id_type
    v.targets[0].id = target
    v.targets[0].data_path = data_path

    d.expression = v.name


def link_driver_multiply(source, target_1, target_2, property, data_path_1, data_path_2, index=-1, id_type_1='WORLD', id_type_2='WORLD'):
    d = source.driver_add(property, index).driver

    for var in d.variables:
        d.variables.remove(var)

    v1 = d.variables.new()
    v1.name = property + "1"
    v1.targets[0].id_type = id_type_1
    v1.targets[0].id = target_1
    v1.targets[0].data_path = data_path_1

    v2 = d.variables.new()
    v2.name = property + "2"
    v2.targets[0].id_type = id_type_2
    v2.targets[0].id = target_2
    v2.targets[0].data_path = data_path_2

    d.expression = v1.name + "*" + v2.name


def get_sun_rotation(sun):
    vec = Vector((0, 0, 1)) # Get sun direction (local Z axis)
    vec.rotate(sun.rotation_euler)

    return vec


def get_previous_node(node):
    for in_socket in node.inputs:
        if in_socket.is_linked:
            for link in in_socket.links:
                if link.is_valid:
                    return link.from_node
    return None


def get_input_node(node, id):
    in_socket = node.inputs[id]
    if in_socket.is_linked:
        for link in in_socket.links:
            if link.is_valid:
                return link.from_node
    return None


def get_next_nodes(node):
    out = []
    for out_socket in node.outputs:
        if out_socket.is_linked:
            for link in out_socket.links:
                if link.is_valid:
                    out.append(link.to_node)
    return out


def move_node_in_between(node_before, the_node, node_after):
    node_padding = 40
    if node_before is None:
        the_node.location.xy = (node_after.location.x - node_after.width - node_padding, node_after.location.y)
    else:
        the_node_xpos = node_before.location.x + node_padding + node_before.width
        node_after_xpos = the_node_xpos + node_padding + node_after.width
        the_node.location.xy = (the_node_xpos, node_before.location.y)
        node_after.location.xy = (node_after_xpos, node_before.location.y)


def put_node_on_link(node_tree, n1, n2, n3):
    node_tree.links.new(n2.outputs[0], n3.inputs[0])
    node_tree.links.new(n1.outputs[0], n2.inputs[0])
    move_node_in_between(n1, n2, n3)

def link_nodes(node_tree, n1, n2):
    link_node_sockets(node_tree, n1, 0, n2, 0)


def link_node_sockets(node_tree, n1, s1, n2, s2):
    if not n1 is None and not n2 is None:
        node_tree.links.new(n1.outputs[s1], n2.inputs[s2])
        #move_node_in_between(None, n1, n2)


def clear_input_node_links(node_tree, node, input_index):
    for link in node.inputs[input_index].links:
        node_tree.links.remove(link)


def unlink_output_sockets(node_tree, node):
    for o in node.outputs:
        for l in o.links:
            node_tree.links.remove(l)


def load_and_create_node_group(group_name, node_name):
    path = os.path.join(os.path.dirname(__file__), "blends/" + PCO_BLEND_FILE + "/NodeTree/")
    bpy.ops.wm.append(filepath=path+node_name, directory=path, filename=node_name, do_reuse_local_id=True)

    if group_name in bpy.data.node_groups:
        group = bpy.data.node_groups[group_name]  # fetch already existing group
        # create and add the new node to the tree
        group_node = bpy.context.scene.world.node_tree.nodes.new(type='ShaderNodeGroup')
        group_node.name = node_name
        group_node.node_tree = group  # assign the existing group

        return group_node
    return None


def get_node_group_from_template(template_name, copy=False, name=''):
    if template_name not in bpy.data.node_groups: # Load template if not present
        path = os.path.join(os.path.dirname(__file__), "blends/" + PCO_BLEND_FILE + "/NodeTree/")
        bpy.ops.wm.append(filepath=path+template_name, directory=path, filename=template_name, do_reuse_local_id=True)
    
    template_group = bpy.data.node_groups.get(template_name)
    if template_group:
        if copy:
            copy = template_group.copy()
            if len(name) > 0:
                copy.name = name
            return copy
        return template_group
    return None


def delete_node_group(node_name, node_tree=None):
    if node_tree is None:
        node_tree = bpy.context.scene.world.node_tree
    node = node_tree.nodes.get(node_name)
    if node is not None:
        post_nodes = get_next_nodes(node)
        pre_node = get_previous_node(node)
        if pre_node is not None:
            for post_node in post_nodes:
                # TODO ensure correct sockets to prevent breakage
                pass
        node_tree.nodes.remove(node)


def remove_named_node_from_group(group, name):
    node = group.nodes.get(name)
    if node:
        group.nodes.remove(node)
