import enum
import bpy
from mathutils import Vector, Euler
import math

from . helpers import *
from . globals import *


def init_sun_data(context):
    world = context.scene.world
    sun = world.pco_general_settings.sun
    input_group = world.pco_exposed.inputs_node

    if sun is not None:
        if sun.type == 'LIGHT' and sun.data.type == 'SUN':
            link_driver_simple(input_group.nodes["sun_dir"].inputs[0], sun, "default_value", "rotation_euler.x", index=0, id_type='OBJECT')
            link_driver_simple(input_group.nodes["sun_dir"].inputs[0], sun, "default_value", "rotation_euler.y", index=1, id_type='OBJECT')
            link_driver_simple(input_group.nodes["sun_dir"].inputs[0], sun, "default_value", "rotation_euler.z", index=2, id_type='OBJECT')

            link_driver_multiply(input_group.nodes["sun_color"].outputs[0], sun, sun, "default_value", "data.color.r", "data.energy", index=0, id_type_1='OBJECT', id_type_2='OBJECT')
            link_driver_multiply(input_group.nodes["sun_color"].outputs[0], sun, sun, "default_value", "data.color.g", "data.energy", index=1, id_type_1='OBJECT', id_type_2='OBJECT')
            link_driver_multiply(input_group.nodes["sun_color"].outputs[0], sun, sun, "default_value", "data.color.b", "data.energy", index=2, id_type_1='OBJECT', id_type_2='OBJECT')

            link_driver_simple(input_group.nodes["sun_angular_diameter"].outputs[0], sun, "default_value", "data.angle", id_type='OBJECT')
        else:
            # TODO Add warning for invalid type
            input_group.nodes["sun_color"].outputs[0].default_value = Vector((0, 0, 0, 0))
    else:
        input_group.nodes["sun_color"].outputs[0].default_value = Vector((0, 0, 0, 0))


def add_planet_transform_nodes(context):
    # One is required for each instance
    # Naming convention is '[NodeName]_[PlanetIndex]'
    # This is duplicating data
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planets = world.pco_planet_slots

    transform_nodes = world.pco_exposed.transform_nodes

    for i in range(len(planets)):
        transform_name = TRANSFORM_NODE_NAME + "_" + str(i)

        transform_group = get_node_group_from_template(
            TRANSFORM_NODE_NAME, copy=True, name=transform_name)

        transform_node = celestial_group.nodes.new(type='ShaderNodeGroup')
        transform_node.name = transform_name
        transform_node.node_tree = transform_group

        reference = transform_nodes.add()
        reference.node_name = transform_name


def add_planet_data_nodes(context):
    # One is required for each instance
    # Naming convention is '[NodeName]_[PlanetIndex]'
    # This is duplicating data
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planets = world.pco_planet_slots

    planet_data_nodes = world.pco_exposed.planet_data_nodes
    material_group_names = world.pco_exposed.material_group_names
    height_normal_group_names = world.pco_exposed.height_normal_group_names

    for i in range(len(planets)):
        planet_data_node_name = PLANET_DATA_NODE_NAME + "_" + str(i)

        planet_data_group = get_node_group_from_template(
            PLANET_DATA_NODE_NAME, copy=True, name=planet_data_node_name)

        planet_data_node = celestial_group.nodes.new(type='ShaderNodeGroup')
        planet_data_node.name = planet_data_node_name
        planet_data_node.node_tree = planet_data_group

        reference = planet_data_nodes.add()
        reference.node_name = planet_data_node_name

        # Ensure unique textured material
        # TODO Ensure this updates based on materiality
        material_node = planet_data_node.node_tree.nodes["textured_material"]
        material_group_name = MATERIAL_GROUP_NAME + "_" + str(i)

        # TODO Currently only textured material exists so no need for extra checks
        material_group = get_node_group_from_template(
            TEXTURED_MATERIAL_DATA_NODE_NAME, copy=True, name=material_group_name)
        material_node.node_tree = material_group

        reference = material_group_names.add()
        reference.node_name = material_group_name

        # This only applies to textured material type at the moment
        height_normal_node = material_node.node_tree.nodes["height_to_normal"]
        height_normal_group_name = HEIGHT_TO_NORMALS_NODE_NAME + "_" + str(i)

        height_normal_group = get_node_group_from_template(
            HEIGHT_TO_NORMALS_NODE_NAME, copy=True, name=height_normal_group_name)
        height_normal_node.node_tree = height_normal_group

        reference = height_normal_group_names.add()
        reference.node_name = height_normal_group_name

        # I am not sure if this is a good solution in terms of maintainability, but it does the job.
        # Worth exploring changing up how the whole stack of textures is handled in the future.
        texture_node_group_name = HEIGHT_TEXTURE_NODE_NAME + "_" + str(i)
        texture_node_group = get_node_group_from_template(
            HEIGHT_TEXTURE_NODE_NAME, copy=True, name=texture_node_group_name)
        for node in height_normal_node.node_tree.nodes:
            if node.name.startswith("TEX_HeightSample"):
                node.node_tree = texture_node_group
        reference = height_normal_group_names.add()
        reference.node_name = texture_node_group_name


def add_planet_depth_sort_nodes(context):
    # One is required for every instance except first
    # Naming convention is '[NodeName]_[PlanetIndex]'
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planets = world.pco_planet_slots

    depth_nodes = world.pco_exposed.depth_nodes

    if len(planets) > 1:
        for i in range(len(planets) - 1):
            depth_name = DEPTH_SORT_NODE_NAME + "_" + str(i)
            depth_group = get_node_group_from_template(DEPTH_SORT_NODE_NAME)

            depth_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            depth_node.name = depth_name
            depth_node.node_tree = depth_group

            reference = depth_nodes.add()
            reference.node_name = depth_name


def add_planet_shadow_nodes(context):
    # One is required for every instance, if more than one present
    # Naming convention is '[NodeName]_[PlanetIndex]'
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planets = world.pco_planet_slots

    shadow_nodes = world.pco_exposed.shadow_nodes

    if len(planets) > 1:
        for i in range(len(planets)):
            shadow_name = PLANET_SHADOW_NODE_NAME + "_" + str(i)
            shadow_group = get_node_group_from_template(PLANET_SHADOW_NODE_NAME)

            shadow_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            shadow_node.name = shadow_name
            shadow_node.node_tree = shadow_group

            reference = shadow_nodes.add()
            reference.node_name = shadow_name


def add_planet_indirect_nodes(context):
    # One is required for every instance, if more than one present
    # Naming convention is '[NodeName]_[PlanetIndex]'
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planets = world.pco_planet_slots

    indirect_nodes = world.pco_exposed.indirect_nodes

    if len(planets) > 1:
        for i in range(len(planets)):
            indirect_name = PLANET_INDIRECT_NODE_NAME + "_" + str(i)
            indirect_group = get_node_group_from_template(PLANET_INDIRECT_NODE_NAME)

            indirect_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            indirect_node.name = indirect_name
            indirect_node.node_tree = indirect_group

            reference = indirect_nodes.add()
            reference.node_name = indirect_name


def add_ring_nodes(context):
    # TODO Change behavior in case we add ring support to multiple planets
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planets = world.pco_planet_slots

    ring_instance = world.pco_exposed.ring_instance

    if not ring_instance < 0:
        instance = planets[ring_instance]
        if not instance.planet_data.rings_type != 'DISABLED':
            ring_instance = -1
        else:
            ring_data_group = get_node_group_from_template(RING_DATA_NODE_NAME)
            ring_data_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            ring_data_node.name = RING_DATA_NODE_NAME
            ring_data_node.node_tree = ring_data_group

            ring_group = get_node_group_from_template(RING_NODE_NAME)
            ring_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            ring_node.name = RING_NODE_NAME
            ring_node.node_tree = ring_group


def add_atmosphere_nodes(context):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planets = world.pco_planet_slots

    ring_instance = world.pco_exposed.ring_instance
    atmosphere_instance = world.pco_exposed.atmosphere_instance

    if not atmosphere_instance < 0:
        instance = planets[atmosphere_instance]
        if not instance.planet_data.atmosphere_type != 'DISABLED':
            atmosphere_instance = -1
        else:
            atmosphere_data_group = get_node_group_from_template(ATMOSPHERE_DATA_NODE_NAME)
            atmosphere_data_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            atmosphere_data_node.name = ATMOSPHERE_DATA_NODE_NAME
            atmosphere_data_node.node_tree = atmosphere_data_group

            atmosphere_node_name = RING_ATMOSPHERE_NODE_NAME
            if ring_instance < 0:
                atmosphere_node_name = ATMOSPHERE_NODE_NAME

            atmosphere_group = get_node_group_from_template(atmosphere_node_name)
            atmosphere_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            atmosphere_node.name = atmosphere_node_name
            atmosphere_node.node_tree = atmosphere_group
    else:
        # We need to blend in rings if no atmosphere is present
        if not ring_instance < 0:
            ring_mask_group = get_node_group_from_template(RINGS_MASK_NODE)
            ring_mask_node = celestial_group.nodes.new(type='ShaderNodeGroup')
            ring_mask_node.name = RINGS_MASK_NODE
            ring_mask_node.node_tree = ring_mask_group


def link_planet_data_nodes(context):
    # Link all inputs for planet data nodes
    # This one is quite simple, just link respective transform node
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    transform_nodes = world.pco_exposed.transform_nodes
    planet_data_nodes = world.pco_exposed.planet_data_nodes

    for i in range(len(transform_nodes)):
        transform_node = celestial_group.nodes.get(transform_nodes[i].node_name)
        planet_data_node = celestial_group.nodes.get(planet_data_nodes[i].node_name)

        link_node_sockets(celestial_group, transform_node, 0, planet_data_node, 0) # Position
        link_node_sockets(celestial_group, transform_node, 1, planet_data_node, 1) # Radius
        link_node_sockets(celestial_group, transform_node, 2, planet_data_node, 2) # Rotation

        planet = world.pco_planet_slots[i]
        clear_input_node_links(celestial_group, transform_node, 8) # Ensure no prior connections to the input
        if planet.transform_mode == 'ORBIT':
            if planet.orbit_parent_type == 'PLANET':
                if not i == planet.orbit_parent_planet:
                    parent_planet_transform_node = celestial_group.nodes.get(
                        transform_nodes[planet.orbit_parent_planet].node_name)
                    link_node_sockets(celestial_group, parent_planet_transform_node, 0, transform_node, 8) # Orbit Origin



def link_depth_sort_nodes(context):
    # Link all inputs for depth sort nodes
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    planet_data_nodes = world.pco_exposed.planet_data_nodes
    depth_nodes = world.pco_exposed.depth_nodes
    
    for i in range(len(depth_nodes)):
        # First node connects two planet data node outputs
        if i == 0:
            planet_data_node_1 = celestial_group.nodes.get(planet_data_nodes[i].node_name)
            planet_data_node_2 = celestial_group.nodes.get(planet_data_nodes[i + 1].node_name)
            depth_sort_node = celestial_group.nodes.get(depth_nodes[i].node_name)

            # First instance
            link_node_sockets(celestial_group, planet_data_node_1, 0, depth_sort_node, 0) # Position
            link_node_sockets(celestial_group, planet_data_node_1, 1, depth_sort_node, 1) # Normal
            link_node_sockets(celestial_group, planet_data_node_1, 2, depth_sort_node, 2) # Albedo
            link_node_sockets(celestial_group, planet_data_node_1, 3, depth_sort_node, 3) # Roughness
            link_node_sockets(celestial_group, planet_data_node_1, 4, depth_sort_node, 4) # Emission
            link_node_sockets(celestial_group, planet_data_node_1, 6, depth_sort_node, 5) # Depth
            link_node_sockets(celestial_group, planet_data_node_1, 7, depth_sort_node, 6) # Alpha
            # Second instance
            link_node_sockets(celestial_group, planet_data_node_2, 0, depth_sort_node, 7) # Position
            link_node_sockets(celestial_group, planet_data_node_2, 1, depth_sort_node, 8) # Normal
            link_node_sockets(celestial_group, planet_data_node_2, 2, depth_sort_node, 9) # Albedo
            link_node_sockets(celestial_group, planet_data_node_2, 3, depth_sort_node, 10) # Roughness
            link_node_sockets(celestial_group, planet_data_node_2, 4, depth_sort_node, 11) # Emission
            link_node_sockets(celestial_group, planet_data_node_2, 6, depth_sort_node, 12) # Depth
            link_node_sockets(celestial_group, planet_data_node_2, 7, depth_sort_node, 13) # Alpha

        # Remaining nodes connect previous node and a planet data node
        else:
            last_depth_sort_node = celestial_group.nodes.get(depth_nodes[i - 1].node_name)
            planet_data_node = celestial_group.nodes.get(planet_data_nodes[i + 1].node_name)
            depth_sort_node = celestial_group.nodes.get(depth_nodes[i].node_name)

            # Previous depth sort node
            link_node_sockets(celestial_group, last_depth_sort_node, 0, depth_sort_node, 0) # Position
            link_node_sockets(celestial_group, last_depth_sort_node, 1, depth_sort_node, 1) # Normal
            link_node_sockets(celestial_group, last_depth_sort_node, 2, depth_sort_node, 2) # Albedo
            link_node_sockets(celestial_group, last_depth_sort_node, 3, depth_sort_node, 3) # Roughness
            link_node_sockets(celestial_group, last_depth_sort_node, 4, depth_sort_node, 4) # Emission
            link_node_sockets(celestial_group, last_depth_sort_node, 5, depth_sort_node, 5) # Depth
            link_node_sockets(celestial_group, last_depth_sort_node, 6, depth_sort_node, 6) # Alpha
            # New instance
            link_node_sockets(celestial_group, planet_data_node, 0, depth_sort_node, 7) # Position
            link_node_sockets(celestial_group, planet_data_node, 1, depth_sort_node, 8) # Normal
            link_node_sockets(celestial_group, planet_data_node, 2, depth_sort_node, 9) # Albedo
            link_node_sockets(celestial_group, planet_data_node, 3, depth_sort_node, 10) # Roughness
            link_node_sockets(celestial_group, planet_data_node, 4, depth_sort_node, 11) # Emission
            link_node_sockets(celestial_group, planet_data_node, 6, depth_sort_node, 12) # Depth
            link_node_sockets(celestial_group, planet_data_node, 7, depth_sort_node, 13) # Alpha


def link_indirect_light_nodes(context):
     # Link all inputs for indirect light nodes
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    transform_nodes = world.pco_exposed.transform_nodes
    planet_data_nodes = world.pco_exposed.planet_data_nodes
    depth_nodes = world.pco_exposed.depth_nodes
    indirect_nodes = world.pco_exposed.indirect_nodes
    celestial_input = celestial_group.nodes.get(CELESTIAL_GROUP_INPUT_NODE)
    
    for i in range(len(indirect_nodes)):
        transform_node = celestial_group.nodes.get(transform_nodes[i].node_name)
        planet_data_node = celestial_group.nodes.get(planet_data_nodes[i].node_name)
        indirect_node = celestial_group.nodes.get(indirect_nodes[i].node_name)

        # We always use the last one for this
        pdn = None
        if len(planet_data_nodes) > 0:
            pdn = celestial_group.nodes.get(planet_data_nodes[0].node_name)
        depth_node = pdn
        if len(depth_nodes) > 0:
            depth_node = celestial_group.nodes.get(depth_nodes[-1].node_name)

        # First node doesn't take input in first slot
        if i > 0:
            last_indirect_node = celestial_group.nodes.get(indirect_nodes[i - 1].node_name)
            link_node_sockets(celestial_group, last_indirect_node, 0, indirect_node, 0) # Indirect Light
        link_node_sockets(celestial_group, planet_data_node, 5, indirect_node, 1) # Caster Average Albedo
        link_node_sockets(celestial_group, depth_node, 0, indirect_node, 2) # Position
        link_node_sockets(celestial_group, depth_node, 1, indirect_node, 3) # Normal
        link_node_sockets(celestial_group, transform_node, 0, indirect_node, 4) # Planet Position (Caster)
        link_node_sockets(celestial_group, transform_node, 1, indirect_node, 5) # Planet Radius (Caster)
        link_node_sockets(celestial_group, celestial_input, 1, indirect_node, 6) # Sun Vector


def link_shadow_nodes(context):
     # Link all inputs for shadow nodes
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    transform_nodes = world.pco_exposed.transform_nodes
    depth_nodes = world.pco_exposed.depth_nodes
    shadow_nodes = world.pco_exposed.shadow_nodes
    celestial_input = celestial_group.nodes.get(CELESTIAL_GROUP_INPUT_NODE)
    
    for i in range(len(shadow_nodes)):
        transform_node = celestial_group.nodes.get(transform_nodes[i].node_name)
        shadow_node = celestial_group.nodes.get(shadow_nodes[i].node_name)

        # We always use the last one for this
        planet_data_node = None
        planet_data_nodes = world.pco_exposed.planet_data_nodes
        if len(planet_data_nodes) > 0:
            planet_data_node = celestial_group.nodes.get(planet_data_nodes[0].node_name)
        depth_node = planet_data_node
        if len(depth_nodes) > 0:
            depth_node = celestial_group.nodes.get(depth_nodes[-1].node_name)

        # First node doesn't take input in first slot
        if i > 0:
            last_shadow_node = celestial_group.nodes.get(shadow_nodes[i - 1].node_name)
            link_node_sockets(celestial_group, last_shadow_node, 0, shadow_node, 0) # Shadow
        link_node_sockets(celestial_group, depth_node, 0, shadow_node, 1) # Position
        link_node_sockets(celestial_group, transform_node, 0, shadow_node, 2) # Planet Position (Caster)
        link_node_sockets(celestial_group, transform_node, 1, shadow_node, 3) # Planet Radius (Caster)
        link_node_sockets(celestial_group, celestial_input, 1, shadow_node, 4) # Sun Vector
        link_node_sockets(celestial_group, celestial_input, 3, shadow_node, 5) # Sun Angular Diameter


def link_ring_node(context):
    # Link all inputs for ring node
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    celestial_input = celestial_group.nodes.get(CELESTIAL_GROUP_INPUT_NODE)

    if not world.pco_exposed.ring_instance < 0:
        transform_node_name = world.pco_exposed.transform_nodes[world.pco_exposed.ring_instance].node_name
        transform_node = celestial_group.nodes.get(transform_node_name)
        ring_data_node = celestial_group.nodes.get(RING_DATA_NODE_NAME)
        ring_node = celestial_group.nodes.get(RING_NODE_NAME)

        link_node_sockets(celestial_group, transform_node, 0, ring_node, 0) # Planet Position
        link_node_sockets(celestial_group, transform_node, 1, ring_node, 1) # Planet Radius
        link_node_sockets(celestial_group, ring_data_node, 0, ring_node, 2) # Ring Vector
        link_node_sockets(celestial_group, ring_data_node, 1, ring_node, 3) # From Min
        link_node_sockets(celestial_group, ring_data_node, 2, ring_node, 4) # From Max
        link_node_sockets(celestial_group, ring_data_node, 3, ring_node, 5) # Indirect Shadow
        link_node_sockets(celestial_group, ring_data_node, 4, ring_node, 6) # Ring Translucency
        link_node_sockets(celestial_group, ring_data_node, 5, ring_node, 7) # Ring Tint
        link_node_sockets(celestial_group, ring_data_node, 6, ring_node, 8) # Ring Texture Scale
        link_node_sockets(celestial_group, ring_data_node, 7, ring_node, 9) # Ring Texture Detail
        link_node_sockets(celestial_group, ring_data_node, 8, ring_node, 10) # Ring Texture Roughness
        link_node_sockets(celestial_group, celestial_input, 1, ring_node, 11) # Sun Vector
        link_node_sockets(celestial_group, celestial_input, 2, ring_node, 12) # Sun Color


def link_atmosphere_node(context):
    # Link all inputs for atmosphere node (including ring-based version), also links atmosphere data node
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    planet_shader_node = celestial_group.nodes.get(PLANET_SHADER_NODE_NAME)
    celestial_input = celestial_group.nodes.get(CELESTIAL_GROUP_INPUT_NODE)

    if not world.pco_exposed.atmosphere_instance < 0:
        transform_node_name = world.pco_exposed.transform_nodes[world.pco_exposed.atmosphere_instance].node_name
        transform_node = celestial_group.nodes.get(transform_node_name)
        atmosphere_data_node = celestial_group.nodes.get(ATMOSPHERE_DATA_NODE_NAME)

        link_node_sockets(celestial_group, transform_node, 1, atmosphere_data_node, 0) # Planet Radius

        if world.pco_exposed.ring_instance < 0:
            # No rings atmosphere
            atmosphere_node = celestial_group.nodes.get(ATMOSPHERE_NODE_NAME)

            link_node_sockets(celestial_group, planet_shader_node, 0, atmosphere_node, 0) # Planet In
            link_node_sockets(celestial_group, celestial_input, 1, atmosphere_node, 1) # Sun Vector
            link_node_sockets(celestial_group, celestial_input, 2, atmosphere_node, 2) # Sun Color
            link_node_sockets(celestial_group, celestial_input, 3, atmosphere_node, 3) # Sun Angular Diameter
            link_node_sockets(celestial_group, transform_node, 1, atmosphere_node, 4) # Planet Radius
            link_node_sockets(celestial_group, transform_node, 0, atmosphere_node, 5) # Planet Position
            link_node_sockets(celestial_group, atmosphere_data_node, 0, atmosphere_node, 6) # Atmosphere Radius
            link_node_sockets(celestial_group, atmosphere_data_node, 1, atmosphere_node, 7) # Rayleigh Scale Height
            link_node_sockets(celestial_group, atmosphere_data_node, 2, atmosphere_node, 8) # Mie Scale Height
            link_node_sockets(celestial_group, atmosphere_data_node, 3, atmosphere_node, 9) # Rayleigh Coefficient
            link_node_sockets(celestial_group, atmosphere_data_node, 4, atmosphere_node, 10) # Mie Coefficient
            link_node_sockets(celestial_group, atmosphere_data_node, 5, atmosphere_node, 11) # Ozone Coefficient
            link_node_sockets(celestial_group, atmosphere_data_node, 6, atmosphere_node, 12) # Density
            link_node_sockets(celestial_group, atmosphere_data_node, 7, atmosphere_node, 13) # Noise Jitter
            link_node_sockets(celestial_group, atmosphere_data_node, 8, atmosphere_node, 14) # Reflection Jitter Multiplier
            link_node_sockets(celestial_group, planet_shader_node, 1, atmosphere_node, 15) # Planet Depth
            link_node_sockets(celestial_group, planet_shader_node, 2, atmosphere_node, 16) # Planet Alpha

        else:
            # Atmosphere with rings
            atmosphere_node = celestial_group.nodes.get(RING_ATMOSPHERE_NODE_NAME)

            ring_data_node = celestial_group.nodes.get(RING_DATA_NODE_NAME)
            ring_node = celestial_group.nodes.get(RING_NODE_NAME)

            link_node_sockets(celestial_group, planet_shader_node, 0, atmosphere_node, 0) # Planet In
            link_node_sockets(celestial_group, celestial_input, 1, atmosphere_node, 1) # Sun Vector
            link_node_sockets(celestial_group, celestial_input, 2, atmosphere_node, 2) # Sun Color
            link_node_sockets(celestial_group, celestial_input, 3, atmosphere_node, 3) # Sun Angular Diameter
            link_node_sockets(celestial_group, transform_node, 1, atmosphere_node, 4) # Planet Radius
            link_node_sockets(celestial_group, transform_node, 0, atmosphere_node, 5) # Planet Position
            link_node_sockets(celestial_group, atmosphere_data_node, 0, atmosphere_node, 6) # Atmosphere Radius
            link_node_sockets(celestial_group, atmosphere_data_node, 1, atmosphere_node, 7) # Rayleigh Scale Height
            link_node_sockets(celestial_group, atmosphere_data_node, 2, atmosphere_node, 8) # Mie Scale Height
            link_node_sockets(celestial_group, atmosphere_data_node, 3, atmosphere_node, 9) # Rayleigh Coefficient
            link_node_sockets(celestial_group, atmosphere_data_node, 4, atmosphere_node, 10) # Mie Coefficient
            link_node_sockets(celestial_group, atmosphere_data_node, 5, atmosphere_node, 11) # Ozone Coefficient
            link_node_sockets(celestial_group, atmosphere_data_node, 6, atmosphere_node, 12) # Density++
            link_node_sockets(celestial_group, atmosphere_data_node, 7, atmosphere_node, 13) # Noise Jitter
            link_node_sockets(celestial_group, atmosphere_data_node, 8, atmosphere_node, 14) # Reflection Jitter Multiplier
            link_node_sockets(celestial_group, ring_node, 0, atmosphere_node, 15) # Rings In
            link_node_sockets(celestial_group, ring_node, 3, atmosphere_node, 16) # Rings Mask
            link_node_sockets(celestial_group, ring_node, 4, atmosphere_node, 17) # Rings Depth
            link_node_sockets(celestial_group, ring_data_node, 0, atmosphere_node, 18) # Ring Vector
            link_node_sockets(celestial_group, ring_node, 5, atmosphere_node, 19) # Ring Radius Min
            link_node_sockets(celestial_group, ring_node, 6, atmosphere_node, 20) # Ring Radius Max
            link_node_sockets(celestial_group, ring_data_node, 3, atmosphere_node, 21) # Ring Indirect Shadow
            link_node_sockets(celestial_group, ring_data_node, 4, atmosphere_node, 22) # Ring Translucency
            link_node_sockets(celestial_group, ring_data_node, 5, atmosphere_node, 23) # Rings Tint
            link_node_sockets(celestial_group, ring_data_node, 6, atmosphere_node, 24) # Ring Texture Scale
            link_node_sockets(celestial_group, ring_data_node, 7, atmosphere_node, 25) # Ring Texture Detail
            link_node_sockets(celestial_group, ring_data_node, 8, atmosphere_node, 26) # Ring Texture Roughness
            link_node_sockets(celestial_group, planet_shader_node, 1, atmosphere_node, 27) # Planet Depth
            link_node_sockets(celestial_group, planet_shader_node, 2, atmosphere_node, 28) # Planet Alpha
    
    else:
        # No atmopsphere, but might need to check for rings.
        if not world.pco_exposed.ring_instance < 0:
            ring_node = celestial_group.nodes.get(RING_NODE_NAME)
            ring_mask_node = celestial_group.nodes.get(RINGS_MASK_NODE)

            link_node_sockets(celestial_group, planet_shader_node, 0, ring_mask_node, 0) # Planet In
            link_node_sockets(celestial_group, ring_node, 0, ring_mask_node, 1) # Rings In
            link_node_sockets(celestial_group, ring_node, 3, ring_mask_node, 2) # Rings Mask


def link_planet_shader_node(context):
    # Link all planet shader node inputs
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    planet_shader_node = celestial_group.nodes.get(PLANET_SHADER_NODE_NAME)
    celestial_input = celestial_group.nodes.get(CELESTIAL_GROUP_INPUT_NODE)

    planet_data_nodes = world.pco_exposed.planet_data_nodes
    depth_nodes = world.pco_exposed.depth_nodes

    planet_data_node = None
    if len(planet_data_nodes) > 0:
        planet_data_node = celestial_group.nodes.get(planet_data_nodes[0].node_name)
    depth_node = planet_data_node
    if len(depth_nodes) > 0:
        depth_node = celestial_group.nodes.get(depth_nodes[-1].node_name)

    indirect_nodes = world.pco_exposed.indirect_nodes
    indirect_node = None
    if len(indirect_nodes) > 0:
        indirect_node = celestial_group.nodes.get(indirect_nodes[-1].node_name)

    shadow_nodes = world.pco_exposed.shadow_nodes
    shadow_node = None
    if len(shadow_nodes) > 0:
        shadow_node = celestial_group.nodes.get(shadow_nodes[-1].node_name)

    link_node_sockets(celestial_group, celestial_input, 0, planet_shader_node, 0) # World
    link_node_sockets(celestial_group, depth_node, 0, planet_shader_node, 1) # Position
    link_node_sockets(celestial_group, depth_node, 1, planet_shader_node, 2) # Normal
    link_node_sockets(celestial_group, depth_node, 2, planet_shader_node, 3) # Albedo
    link_node_sockets(celestial_group, depth_node, 3, planet_shader_node, 4) # Roughness
    link_node_sockets(celestial_group, depth_node, 4, planet_shader_node, 5) # Emission
    link_node_sockets(celestial_group, shadow_node, 0, planet_shader_node, 6) # Shadow
    link_node_sockets(celestial_group, indirect_node, 0, planet_shader_node, 8) # Indirect Light
    link_node_sockets(celestial_group, celestial_input, 1, planet_shader_node, 10) # Sun Vector
    link_node_sockets(celestial_group, celestial_input, 2, planet_shader_node, 11) # Sun Color
    link_node_sockets(celestial_group, celestial_input, 3, planet_shader_node, 12) # Sun Angular Diameter
    # Single planet data node links differently
    if len(planet_data_nodes) == 1:
        link_node_sockets(celestial_group, depth_node, 6, planet_shader_node, 13) # Depth
        link_node_sockets(celestial_group, depth_node, 7, planet_shader_node, 14) # Alpha
    else:
        link_node_sockets(celestial_group, depth_node, 5, planet_shader_node, 13) # Depth
        link_node_sockets(celestial_group, depth_node, 6, planet_shader_node, 14) # Alpha

    if not world.pco_exposed.ring_instance < 0:
        ring_node = celestial_group.nodes.get(RING_NODE_NAME)
        link_node_sockets(celestial_group, ring_node, 2, planet_shader_node, 7) # Rings Shadow
        link_node_sockets(celestial_group, ring_node, 1, planet_shader_node, 9) # Rings Indirect Light


def link_celestial_output_node(context):
    # Link the correct color output for the whole node group
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    celestial_output = celestial_group.nodes.get(CELESTIAL_GROUP_OUTPUT_NODE)

    if world.pco_exposed.atmosphere_instance < 0:
        if world.pco_exposed.ring_instance < 0:
            planet_shader_node = celestial_group.nodes.get(PLANET_SHADER_NODE_NAME)
            link_node_sockets(celestial_group, planet_shader_node, 0, celestial_output, 0)
        else:
            ring_mask_node = celestial_group.nodes.get(RINGS_MASK_NODE)
            link_node_sockets(celestial_group, ring_mask_node, 0, celestial_output, 0)
    else:
        # Output is same for both so can access this way
        atmosphere_node = celestial_group.nodes.get(RING_ATMOSPHERE_NODE_NAME)
        if world.pco_exposed.ring_instance < 0:
            atmosphere_node = celestial_group.nodes.get(ATMOSPHERE_NODE_NAME)
        link_node_sockets(celestial_group, atmosphere_node, 0, celestial_output, 0)


def organize_nodes(context):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    HORIZONTAL_PADDING = 250
    MINI_VERTICAL_PADDING = 350
    VERTICAL_PADDING = 600

    depth_nodes = world.pco_exposed.depth_nodes
    shadow_nodes = world.pco_exposed.shadow_nodes
    indirect_nodes = world.pco_exposed.indirect_nodes
    transform_nodes = world.pco_exposed.transform_nodes
    planet_data_nodes = world.pco_exposed.planet_data_nodes

    planet_shader_node = celestial_group.nodes.get(PLANET_SHADER_NODE_NAME)
    celestial_input = celestial_group.nodes.get(CELESTIAL_GROUP_INPUT_NODE)
    celestial_output = celestial_group.nodes.get(CELESTIAL_GROUP_OUTPUT_NODE)

    planet_shader_node.location.xy = (0, 0)

    celestial_input.location.xy = (-HORIZONTAL_PADDING * 3, MINI_VERTICAL_PADDING)
    celestial_output.location.xy = (HORIZONTAL_PADDING * 2, 0)

    for i in range(len(depth_nodes)):
        depth_node = celestial_group.nodes.get(depth_nodes[i].node_name)
        depth_node.location.xy = (-HORIZONTAL_PADDING * (len(depth_nodes) - i), 0)
    
    for i in range(len(shadow_nodes)):
        shadow_node = celestial_group.nodes.get(shadow_nodes[i].node_name)
        shadow_node.location.xy = (-HORIZONTAL_PADDING * (len(shadow_nodes) - i), -VERTICAL_PADDING)

    for i in range(len(indirect_nodes)):
        indirect_node = celestial_group.nodes.get(indirect_nodes[i].node_name)
        indirect_node.location.xy = (-HORIZONTAL_PADDING * (len(indirect_nodes) - i), -VERTICAL_PADDING - MINI_VERTICAL_PADDING)
    
    for i in range(len(planet_data_nodes)):
        planet_data_node = celestial_group.nodes.get(planet_data_nodes[i].node_name)
        transform_node = celestial_group.nodes.get(transform_nodes[i].node_name)
        planet_data_node.location.xy = (-HORIZONTAL_PADDING * (len(shadow_nodes) + 1), -MINI_VERTICAL_PADDING * i)
        transform_node.location.xy = (-HORIZONTAL_PADDING * (len(shadow_nodes) + 2), -MINI_VERTICAL_PADDING * i)


    atmosphere_node = celestial_group.nodes.get(ATMOSPHERE_NODE_NAME)
    if atmosphere_node:
        atmosphere_node.location.xy = (HORIZONTAL_PADDING, 0)

    ring_atmosphere_node = celestial_group.nodes.get(RING_ATMOSPHERE_NODE_NAME)
    if ring_atmosphere_node:
        ring_atmosphere_node.location.xy = (HORIZONTAL_PADDING, 0)

    atmosphere_data_node = celestial_group.nodes.get(ATMOSPHERE_DATA_NODE_NAME)
    if atmosphere_data_node:
        atmosphere_data_node.location.xy = (0, -VERTICAL_PADDING)
    
    ring_mask_node = celestial_group.nodes.get(RINGS_MASK_NODE)
    if ring_mask_node:
        ring_mask_node.location.xy = (HORIZONTAL_PADDING, 0)

    ring_node = celestial_group.nodes.get(RING_NODE_NAME)
    if ring_node:
        ring_node.location.xy = (HORIZONTAL_PADDING, -VERTICAL_PADDING - MINI_VERTICAL_PADDING)
    
    ring_data_node = celestial_group.nodes.get(RING_DATA_NODE_NAME)
    if ring_data_node:
        ring_data_node.location.xy = (0, -VERTICAL_PADDING - MINI_VERTICAL_PADDING)


def add_transform_drivers(context):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    transform_nodes = world.pco_exposed.transform_nodes

    for i in range(len(world.pco_planet_slots)):
        transform_node = celestial_group.nodes.get(transform_nodes[i].node_name)

        prop_path = "pco_planet_slots[" + str(i) + "]."

        drivers = (
            ["Radius", prop_path + "radius", -1],
            ["Rotation", prop_path + "rotation.x", 0],
            ["Rotation", prop_path + "rotation.y", 1],
            ["Rotation", prop_path + "rotation.z", 2],
            ["Position", prop_path + "position.x", 0],
            ["Position", prop_path + "position.y", 1],
            ["Position", prop_path + "position.z", 2],
            ["Distance From Viewer", prop_path + "distance", -1],
            ["Elevation", prop_path + "elevation", -1],
            ["Azimuth", prop_path + "azimuth", -1],
            ["Orbit Rotation", prop_path + "orbit_rotation.x", 0],
            ["Orbit Rotation", prop_path + "orbit_rotation.y", 1],
            ["Orbit Rotation", prop_path + "orbit_rotation.z", 2],
            ["Ellipse Width", prop_path + "ellipse_width", -1],
            ["Ellipse Height", prop_path + "ellipse_height", -1],
            ["Angle on Ellipse", prop_path + "orbit_point_angle", -1],
        )

        planet = world.pco_planet_slots[i]

        transform_node.inputs["Use View Transform"].default_value = 0.0
        transform_node.inputs["Use Orbital Transform"].default_value = 0.0
        if planet.transform_mode == 'VIEWER':
            transform_node.inputs["Use View Transform"].default_value = 1.0
        if planet.transform_mode == 'ORBIT':
            transform_node.inputs["Use Orbital Transform"].default_value = 1.0

            if planet.orbit_parent_type == 'POINT':
                drivers = drivers + (
                    ["Orbit Origin", prop_path + "orbit_parent_point.x", 0],
                    ["Orbit Origin", prop_path + "orbit_parent_point.y", 1],
                    ["Orbit Origin", prop_path + "orbit_parent_point.z", 2],
                )
            elif planet.orbit_parent_type == 'OBJECT':
                drivers = drivers + (
                    ["Orbit Origin", prop_path + "orbit_parent_object.location.x", 0],
                    ["Orbit Origin", prop_path + "orbit_parent_object.location.y", 1],
                    ["Orbit Origin", prop_path + "orbit_parent_object.location.z", 2],
                )

        for data in drivers:
            link_driver_simple(transform_node.inputs[data[0]], world, "default_value", data[1], index=data[2], id_type='WORLD')


def add_textured_material_drivers(context):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    planet_data_nodes = world.pco_exposed.planet_data_nodes

    for i in range(len(world.pco_planet_slots)):
        planet_data_node = celestial_group.nodes.get(planet_data_nodes[i].node_name)
        material_node = planet_data_node.node_tree.nodes.get("textured_material")

        prop_path = "pco_planet_slots[" + str(i) + "].planet_data."

        drivers = (
            ["Albedo", prop_path + "albedo.r", 0],
            ["Albedo", prop_path + "albedo.g", 1],
            ["Albedo", prop_path + "albedo.b", 2],
            ["Height Scale", prop_path + "height_scale", -1],
            ["Roughness", prop_path + "roughness", -1],
            ["Emission", prop_path + "emission.r", 0],
            ["Emission", prop_path + "emission.g", 1],
            ["Emission", prop_path + "emission.b", 2],
            ["Emission Strength", prop_path + "emission_strength", -1],
        )

        for data in drivers:
            link_driver_simple(material_node.inputs[data[0]], world, "default_value", data[1], index=data[2], id_type='WORLD')


def add_atmosphere_drivers(context):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    atmosphere_data_node = celestial_group.nodes.get(ATMOSPHERE_DATA_NODE_NAME)

    if atmosphere_data_node and not world.pco_exposed.atmosphere_instance < 0:
        prop_path = "pco_planet_slots[" + str(world.pco_exposed.atmosphere_instance) + "].planet_data."

        drivers = (
            ["Radius", prop_path + "atmosphere_radius", -1],
            ["Rayleigh Scale Height", prop_path + "atmosphere_rayleigh_scale_height", -1],
            ["Rayleight Coefficient", prop_path + "atmosphere_rayleigh_coefficient.r", 0],
            ["Rayleight Coefficient", prop_path + "atmosphere_rayleigh_coefficient.g", 1],
            ["Rayleight Coefficient", prop_path + "atmosphere_rayleigh_coefficient.b", 2],
            ["Mie Scale Height", prop_path + "atmosphere_mie_scale_height", -1],
            ["Mie Coefficient", prop_path + "atmosphere_mie_coefficient.r", 0],
            ["Mie Coefficient", prop_path + "atmosphere_mie_coefficient.g", 1],
            ["Mie Coefficient", prop_path + "atmosphere_mie_coefficient.b", 2],
            ["Density", prop_path + "atmosphere_density", -1],
            ["Noise Jitter", prop_path + "atmosphere_noise_jitter", -1],
            ["Reflection Jitter Multiplier", prop_path + "atmosphere_reflection_jitter_multiplier", -1],
        )

        for data in drivers:
            link_driver_simple(atmosphere_data_node.inputs[data[0]], world, "default_value", data[1], index=data[2], id_type='WORLD')


def add_ring_drivers(context):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    ring_data_node = celestial_group.nodes.get(RING_DATA_NODE_NAME)

    if ring_data_node and not world.pco_exposed.ring_instance < 0:
        prop_path = "pco_planet_slots[" + str(world.pco_exposed.ring_instance) + "].planet_data."

        drivers = (
            ["Ring Vector", prop_path + "rings_vector.x", 0],
            ["Ring Vector", prop_path + "rings_vector.y", 1],
            ["Ring Vector", prop_path + "rings_vector.z", 2],
            ["Rings Start", prop_path + "rings_start", -1],
            ["Rings End", prop_path + "rings_end", -1],
            ["Indirect Shadow", prop_path + "rings_indirect_shadow", -1],
            ["Rings Translucency", prop_path + "rings_translucency", -1],
            ["Rings Tint", prop_path + "rings_tint.r", 0],
            ["Rings Tint", prop_path + "rings_tint.g", 1],
            ["Rings Tint", prop_path + "rings_tint.b", 2],
            ["Ring Texture Scale", prop_path + "ring_procedural_texture_scale", -1],
            ["Ring Texture Detail", prop_path + "ring_procedural_texture_detail", -1],
            ["Ring Texture Roughness", prop_path + "ring_procedural_texture_roughness", -1],
        )

        for data in drivers:
            link_driver_simple(ring_data_node.inputs[data[0]], world, "default_value", data[1], index=data[2], id_type='WORLD')


def set_all_textures(context):
    world = context.scene.world
    for i in range(len(world.pco_planet_slots)):
        set_albedo_texture(context, i)
        set_roughness_texture(context, i)
        set_emission_texture(context, i)
        set_height_texture(context, i)


def set_albedo_texture(context, index):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planet = world.pco_planet_slots[index]

    planet_data_node_name = world.pco_exposed.planet_data_nodes[index].node_name
    planet_data_node = celestial_group.nodes.get(planet_data_node_name)
    material_node = planet_data_node.node_tree.nodes.get("textured_material")
    albedo_texture_node = material_node.node_tree.nodes.get("albedo_texture")

    if planet.planet_data.albedo_texture:
        albedo_texture_node.image = planet.planet_data.albedo_texture
        material_node.inputs["Use Albedo Color"].default_value = 0.0
    else:
        albedo_texture_node.image = None
        material_node.inputs["Use Albedo Color"].default_value = 1.0


def set_roughness_texture(context, index):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planet = world.pco_planet_slots[index]

    planet_data_node_name = world.pco_exposed.planet_data_nodes[index].node_name
    planet_data_node = celestial_group.nodes.get(planet_data_node_name)
    material_node = planet_data_node.node_tree.nodes.get("textured_material")
    roughness_texture_node = material_node.node_tree.nodes.get("roughness_texture")

    if planet.planet_data.roughness_texture:
        roughness_texture_node.image = planet.planet_data.roughness_texture
        material_node.inputs["Use Roughness Value"].default_value = 0.0
    else:
        roughness_texture_node.image = None
        material_node.inputs["Use Roughness Value"].default_value = 1.0


def set_height_texture(context, index):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planet = world.pco_planet_slots[index]

    planet_data_node_name = world.pco_exposed.planet_data_nodes[index].node_name
    planet_data_node = celestial_group.nodes.get(planet_data_node_name)
    material_node = planet_data_node.node_tree.nodes.get("textured_material")
    height_texture_node_group = material_node.node_tree.nodes.get("height_to_normal")
    height_texture_sample = height_texture_node_group.node_tree.nodes.get("TEX_HeightSample1")
    height_texture_node = height_texture_sample.node_tree.nodes.get("Texture")

    if planet.planet_data.height_texture:
        height_texture_node.image = planet.planet_data.height_texture
        material_node.inputs["Use Height Texture"].default_value = 1.0
        material_node.inputs["Height Texture Size"].default_value = max(
            planet.planet_data.height_texture.size[0], planet.planet_data.height_texture.size[1])
    else:
        height_texture_node.image = None
        material_node.inputs["Use Height Texture"].default_value = 0.0


def set_emission_texture(context, index):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node
    planet = world.pco_planet_slots[index]

    planet_data_node_name = world.pco_exposed.planet_data_nodes[index].node_name
    planet_data_node = celestial_group.nodes.get(planet_data_node_name)
    material_node = planet_data_node.node_tree.nodes.get("textured_material")
    emission_texture_node = material_node.node_tree.nodes.get("emission_texture")

    if planet.planet_data.emission_texture:
        emission_texture_node.image = planet.planet_data.emission_texture
        material_node.inputs["Use Emission Color"].default_value = 0.0
    else:
        emission_texture_node.image = None
        material_node.inputs["Use Emission Color"].default_value = 1.0


def set_ring_texture(context, index):
    world = context.scene.world
    celesial_group = world.pco_exposed.celestial_node
    planet_data = world.pco_planet_slots[index].planet_data

    if index == world.pco_exposed.ring_instance:
        ring_node = celesial_group.nodes.get(RING_NODE_NAME)
        if ring_node:
            ring_texture_sample = ring_node.node_tree.nodes.get("TEX_RingTexture")
            ring_texture_node = ring_texture_sample.node_tree.nodes.get("RING_TEXTURE")
            output_node = ring_texture_sample.node_tree.nodes.get("RING_TEXTURE_OUTPUT")

            if planet_data.rings_type == 'TEXTURE':
                link_node_sockets(ring_texture_sample.node_tree, ring_texture_node, 0, output_node, 0) # Color
                link_node_sockets(ring_texture_sample.node_tree, ring_texture_node, 1, output_node, 1) # Alpha
            elif planet_data.rings_type == 'PROCEDURAL':
                procedural_texture_node = ring_texture_sample.node_tree.nodes.get("RING_PROCEDURAL_TEXTURE")
                procedural_alpha_node = ring_texture_sample.node_tree.nodes.get("RING_PROCEDURAL_ALPHA")
                link_node_sockets(ring_texture_sample.node_tree, procedural_texture_node, 0, output_node, 0) # Color
                link_node_sockets(ring_texture_sample.node_tree, procedural_alpha_node, 0, output_node, 1) # Alpha

            ring_texture_node.image = planet_data.rings_texture


def cleanup_celestial_node_tree(context):
    world = context.scene.world
    celestial_group = world.pco_exposed.celestial_node

    depth_nodes = world.pco_exposed.depth_nodes
    shadow_nodes = world.pco_exposed.shadow_nodes
    indirect_nodes = world.pco_exposed.indirect_nodes
    transform_nodes = world.pco_exposed.transform_nodes
    planet_data_nodes = world.pco_exposed.planet_data_nodes
    material_group_names = world.pco_exposed.material_group_names
    height_normal_group_names = world.pco_exposed.height_normal_group_names

    for d_n in depth_nodes:
        celestial_group.nodes.remove(celestial_group.nodes.get(d_n.node_name))
    depth_nodes.clear()

    for s_n in shadow_nodes:
        celestial_group.nodes.remove(celestial_group.nodes.get(s_n.node_name))
    shadow_nodes.clear()

    for i_n in indirect_nodes:
        celestial_group.nodes.remove(celestial_group.nodes.get(i_n.node_name))
    indirect_nodes.clear()

    for t_n in transform_nodes:
        bpy.data.node_groups.remove(bpy.data.node_groups.get(t_n.node_name))
        celestial_group.nodes.remove(celestial_group.nodes.get(t_n.node_name))
    transform_nodes.clear()

    for p_n in planet_data_nodes:
        bpy.data.node_groups.remove(bpy.data.node_groups.get(p_n.node_name))
        celestial_group.nodes.remove(celestial_group.nodes.get(p_n.node_name))
    planet_data_nodes.clear()

    for m_g in material_group_names:
        bpy.data.node_groups.remove(bpy.data.node_groups.get(m_g.node_name))
    material_group_names.clear()

    for h_g in height_normal_group_names:
        bpy.data.node_groups.remove(bpy.data.node_groups.get(h_g.node_name))
    height_normal_group_names.clear()

    remove_named_node_from_group(celestial_group, ATMOSPHERE_NODE_NAME)
    remove_named_node_from_group(celestial_group, RING_ATMOSPHERE_NODE_NAME)
    remove_named_node_from_group(celestial_group, ATMOSPHERE_DATA_NODE_NAME)
    remove_named_node_from_group(celestial_group, RING_NODE_NAME)
    remove_named_node_from_group(celestial_group, RING_DATA_NODE_NAME)
    remove_named_node_from_group(celestial_group, RINGS_MASK_NODE)


def add_planet_instance(context):
    list = context.scene.world.pco_planet_slots

    instance = list.add()
    instance.name = "Planet" + str(len(list))

    return instance


def remove_planet_instance(context, index):
    list = context.scene.world.pco_planet_slots
    planet = list[index]
    list.remove(index)

