import bpy
from bpy.types import (
    Panel,
    UIList,
    Operator,
)

from bpy.props import *

from bpy_extras.node_utils import find_node_input

from . planets import *
from . handlers import update_node_tree, reinitialize_drivers


ADDON_NAME = 'physical-celestial-objects'


class PA_UL_PlanetList(UIList):
    """A UIList for planet instances"""

    def draw_item(self, context, layout, data, item, icon, active_data,
                  active_propname, index):

        custom_icon = 'WORLD_DATA' # TODO Proper icons

        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            layout.prop(item, "name", expand=False, emboss=False, icon = custom_icon)

        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.label(text="", icon = custom_icon)



class PA_OT_ReinitializeDrivers(Operator):
    """Reinitialize all drivers in the node tree"""

    bl_idname = "pco.reinitialize_drivers"
    bl_label = "Reinitialize Drivers"

    def execute(self, context):
        reinitialize_drivers(context)

        return {'FINISHED'}


class PA_OT_NewPlanetInstance(Operator):
    """Add a new planet instance to the list"""

    bl_idname = "pco_planet_slots.new_planet_instance"
    bl_label = "Add a new planet instance"

    def execute(self, context):
        world = context.scene.world
        index = world.pco_active_planet_index

        planet = add_planet_instance(context)

        # Select newly added instance (which is last index)
        context.scene.world.pco_active_planet_index = len(world.pco_planet_slots) - 1

        update_node_tree(context)

        return{'FINISHED'}


class PA_OT_RemovePlanetInstance(Operator):
    """Remove selected planet instance from the list"""

    bl_idname = "pco_planet_slots.remove_planet_instance"
    bl_label = "Remove selected planet instance"

    @classmethod
    def poll(cls, context):
        return context.scene.world.pco_planet_slots
    
    def execute(self, context):
        world = context.scene.world
        index = world.pco_active_planet_index
        remove_planet_instance(context, index)

        # Ensure stability on atmosphere/rings
        if world.pco_exposed.atmosphere_instance == index:
            world.pco_exposed.atmosphere_instance = -1
        elif world.pco_exposed.atmosphere_instance < index:
            world.pco_exposed.atmosphere_instance = max(world.pco_exposed.atmosphere_instance - 1, -1)
        
        if world.pco_exposed.ring_instance == index:
            world.pco_exposed.ring_instance = -1
        elif world.pco_exposed.ring_instance < index:
            world.pco_exposed.ring_instance = max(world.pco_exposed.ring_instance - 1, -1)

        context.scene.world.pco_active_planet_index = min(
            max(0, index - 1), len(world.pco_planet_slots) - 1) # Ensure a valid index

        update_node_tree(context)

        return{'FINISHED'}


class PA_OT_MovePlanetInstance(Operator):
    """Move a a planet instance within the list"""

    bl_idname = "pco_planet_slots.move_planet"
    bl_label = "Move a planet instance in list"

    direction: bpy.props.EnumProperty(items=(('UP', 'Up', ""), ('DOWN', 'Down', ""),))

    @classmethod
    def poll(cls, context):
        return context.scene.world.pco_planet_slots

    def move_index(self):
        """ Move index of an item render queue while clamping it."""
        index = bpy.context.scene.world.pco_active_planet_index
        list_length = len(bpy.context.scene.world.pco_planet_slots) - 1
        new_index = index + (-1 if self.direction == 'UP' else 1)

        bpy.context.scene.world.pco_active_planet_index = max(0, min(new_index, list_length))

    def execute(self, context):
        list = bpy.context.scene.world.pco_planet_slots
        index = bpy.context.scene.world.pco_active_planet_index

        neighbor = index + (-1 if self.direction == 'UP' else 1)
        list.move(neighbor, index)
        self.move_index()

        return{'FINISHED'}


class PA_OT_ChoosePlanetInstance(Operator):
    bl_idname = "pco.choose_planet_instance"
    bl_label = "Choose planet instance from list"
    bl_options = {'INTERNAL'}
    bl_property = "enum"

    def get_enum_from_data(self, context):
        pdata = bpy.data.pco_planets
        new_enum = []
        for p in pdata:
            new_enum.append((p.name))
        return new_enum

    enum: EnumProperty(items=get_enum_from_data, name="Items")

    def execute(self, context):
        world = context.scene.world
        active_instance = world.pco_planet_slots[world.pco_active_planet_index]
        #active_instance.
        return{'FINISHED'}

    def invoke(self, context):
        context.window_manager.invoke_search_popup(self)
        return{'FINISHED'}


class PA_OT_NewPlanetData(Operator):
    bl_idname = "pco.new_planet_data"
    bl_label = "Create planet data"
    bl_options = {'INTERNAL'}

    def execute(self, context):
        new_data = context.scene.pco_planets.add()

        world = context.scene.world
        active_instance = world.pco_planet_slots[world.pco_active_planet_index]

        active_instance.planet_data = new_data
        new_data.node_tree = world.node_tree.nodes[active_instance.planet_data_node].node_tree

        return{'FINISHED'}


class PA_PCO_TB:
    bl_parent_id = "PA_PT_CelestialObjectsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"


# Base class
class DrawablePanel:
    @classmethod
    def poll(self, context):
        settings_available = context.scene and hasattr(context.scene.world, 'pco_general_settings')
        if not settings_available:
            return False
        # toolbar preferences are stored under preferences Class.
        #addon_prefs = context.preferences.addons[ADDON_NAME].preferences
        #if self.bl_context == "objectmode" and not addon_prefs.toolbar_enabled:  # if drawing toolbar but it's not enabled
        #    return False
        return True

    def draw_header(self, context):
        self.layout.label(text='', icon=self.icon_name)

    # indented column with a padding from left
    def indentedColumn(self, squished=True):
        row = self.layout.row()
        row.separator()
        return row.column(align=squished)


class CelestialObjects(DrawablePanel):
    bl_label = "Physical Celestial Objects"

    @classmethod
    def poll(self, context):
        return True

    def draw_header(self, context):
        if context.scene and hasattr(context.scene.world, 'pco_general_settings'):
            settings = context.scene.world.pco_general_settings
            layout = self.layout
            layout.prop(settings, 'enabled', text='')

    def draw(self, context):
        layout = self.layout

        if context.scene:
            layout.template_ID(context.scene, "world", new="world.new")
            if context.scene.world is None:
                layout.label(text="A world is required by Physical Celestial Objects to work.", icon='ERROR')
            else:
                pco_gsettings = context.scene.world.pco_general_settings
                layout.enabled = pco_gsettings.enabled

        # Standalone settings
        #if pco_gsettings.use_standalone_lighting:
        col = layout.column(align=True)
        col.prop(pco_gsettings, 'sun', icon='LIGHT_SUN')
        # TODO Add warning for wrong object type

        # Planet configuration
        # Taken from Blender's material UI
        world = context.scene.world
        planet_slots = world.pco_planet_slots

        if world:
            is_sortable = len(planet_slots) > 1
            rows = 3
            if is_sortable:
                rows = 5

            layout.label(text="Planets:")
            layout.operator("pco.reinitialize_drivers", icon='DRIVER', text="Reinitialize Drivers")
            layout.operator("pco_planet_slots.new_planet_instance", icon='WORLD_DATA', text="Add New Planet")
            row = layout.row()

            row.template_list("PA_UL_PlanetList", "", world, "pco_planet_slots", world, "pco_active_planet_index", rows=rows)

            col = row.column(align=True)
            col.operator("pco_planet_slots.new_planet_instance", icon='ADD', text="")
            col.operator("pco_planet_slots.remove_planet_instance", icon='REMOVE', text="")

            col.separator()

            # col.menu("PCO_MT_context_menu", icon='DOWNARROW_HLT', text="")

            if is_sortable:
                col.separator()
                col.operator("pco_planet_slots.move_planet", icon='TRIA_UP', text="").direction = 'UP'
                col.operator("pco_planet_slots.move_planet", icon='TRIA_DOWN', text="").direction = 'DOWN'


class PA_PT_CelestialObjectsTB(Panel, CelestialObjects):
    bl_idname = "PA_PT_CelestialObjectsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "PCO"


############################################################################
# Transform
############################################################################
class PlanetTransform(DrawablePanel):
    bl_label = "Transform"
    icon_name = "ORIENTATION_GLOBAL"
    
    def draw(self, context):
        if context.scene and hasattr(context.scene.world, 'pco_general_settings'):
            world = context.scene.world
            if len(world.pco_planet_slots) > 0:
                layout = self.layout
                layout.enabled = True

                planet_slots = world.pco_planet_slots
                planet_index = world.pco_active_planet_index
                planet_instance = planet_slots[planet_index]

                layout.prop(planet_instance, "transform_mode", expand=True)

                col = self.indentedColumn()
                col.prop(planet_instance, "radius")
                col.prop(planet_instance, "rotation")
                col.separator()

                if planet_instance.transform_mode == 'VIEWER':
                    col.separator()
                    col.prop(planet_instance, "azimuth")
                    col.prop(planet_instance, "elevation")
                    col.separator()
                    col.prop(planet_instance, "distance")
                elif planet_instance.transform_mode == 'ORBIT':
                    col.separator()
                    col.prop(planet_instance, "orbit_parent_type")
                    if planet_instance.orbit_parent_type == 'OBJECT':
                        col.prop(planet_instance, "orbit_parent_object")
                    elif planet_instance.orbit_parent_type == 'PLANET':
                        col.template_list("PA_UL_PlanetList", "", world, "pco_planet_slots", planet_instance, "orbit_parent_planet", rows=2)
                    else:
                        col.prop(planet_instance, "orbit_parent_point")
                    col.separator()
                    col.prop(planet_instance, "ellipse_width")
                    col.prop(planet_instance, "ellipse_height")
                    col.separator()
                    col.prop(planet_instance, "orbit_rotation")
                    col.separator()
                    col.prop(planet_instance, "orbit_point_angle")
                else:
                    # Assume 'GLOBAL' by default
                    col.prop(planet_instance, "position")

        

class PA_PT_PlanetTransformTB(PA_PCO_TB, Panel, PlanetTransform):
    pass


############################################################################
# Planet Data
############################################################################
class PlanetData(DrawablePanel):
    bl_label = "Surface Material"
    icon_name = "MATERIAL"

    def draw(self, context):
        if context.scene and hasattr(context.scene.world, 'pco_general_settings'):
            world = context.scene.world
            if len(world.pco_planet_slots) > 0:
                layout = self.layout
                layout.enabled = True

                planet_slots = world.pco_planet_slots
                planet_index = world.pco_active_planet_index
                planet_instance = planet_slots[planet_index]

                planet_data = planet_instance.planet_data

                col = self.indentedColumn()

                # if planet_data.material_type == 'TEXTURED':

                col.label(text="Albedo")
                box = col.box()
                box.template_ID(planet_data, "albedo_texture", new="image.new", open="image.open")
                if planet_data.albedo_texture is not None:
                    box.prop(planet_data.albedo_texture.colorspace_settings, "name", text="")
                #if planet_data.albedo_texture is None:
                col.prop(planet_data, "albedo", text="")
                col.separator()

                col.label(text="Roughness")
                box = col.box()
                box.template_ID(planet_data, "roughness_texture", new="image.new", open="image.open")
                if planet_data.roughness_texture is not None:
                    box.prop(planet_data.roughness_texture.colorspace_settings, "name", text="")
                if planet_data.roughness_texture is None:
                    col.prop(planet_data, "roughness", text="Roughness")
                col.separator()

                col.label(text="Height")
                box = col.box()
                box.template_ID(planet_data, "height_texture", new="image.new", open="image.open")
                if planet_data.height_texture is not None:
                    box.prop(planet_data.height_texture.colorspace_settings, "name", text="")
                if planet_data.height_texture is not None:
                    col.prop(planet_data, "height_scale")
                col.separator()

                col.label(text="Emission")
                box = col.box()
                box.template_ID(planet_data, "emission_texture", new="image.new", open="image.open")
                if planet_data.emission_texture is not None:
                    box.prop(planet_data.emission_texture.colorspace_settings, "name", text="")
                if planet_data.emission_texture is None:
                    col.prop(planet_data, "emission", text="")
                col.prop(planet_data, "emission_strength", text="Emission strength")
                col.separator()


class PA_PT_PlanetDataTB(PA_PCO_TB, Panel, PlanetData):
    pass


class PlanetAtmosphere(DrawablePanel):
    bl_label = "Atmosphere"
    icon_name = "PROP_ON"

    def draw(self, context):
        if context.scene and hasattr(context.scene.world, 'pco_general_settings'):
            world = context.scene.world
            if len(world.pco_planet_slots) > 0:
                planet_slots = world.pco_planet_slots
                planet_index = world.pco_active_planet_index
                planet_instance = planet_slots[planet_index]

                planet_data = planet_instance.planet_data

                layout = self.layout

                if world.pco_exposed.atmosphere_instance < 0 or world.pco_exposed.atmosphere_instance == planet_index:
                    layout.prop(planet_data, "atmosphere_type", expand=True)
                    col = self.indentedColumn()
                    if planet_data.atmosphere_type == 'ENABLED':
                        col.prop(planet_data, "atmosphere_radius")
                        col.prop(planet_data, "atmosphere_density")
                        col.prop(planet_data, "atmosphere_rayleigh_coefficient")
                        col.prop(planet_data, "atmosphere_rayleigh_scale_height")
                        col.prop(planet_data, "atmosphere_mie_coefficient")
                        col.prop(planet_data, "atmosphere_mie_scale_height")
                        col.prop(planet_data, "atmosphere_noise_jitter")
                        col.prop(planet_data, "atmosphere_reflection_jitter_multiplier")
                        col.separator()
                elif world.pco_exposed.atmosphere_instance < len(world.pco_planet_slots):
                    layout.scale_y = 0.6
                    layout.label(text='Atmosphere is only supported', icon='ERROR')
                    warning_lines = [
                        "for one planet at a time! Currently",
                        "enabled on: " + planet_slots[world.pco_exposed.atmosphere_instance].name,
                    ]
                    for line in warning_lines:
                        layout.label(text=line)


class PA_PT_PlanetAtmosphereTB(PA_PCO_TB, Panel, PlanetAtmosphere):
    pass


class PlanetRings(DrawablePanel):
    bl_label = "Rings"
    icon_name = "PROP_OFF"

    def draw(self, context):
        if context.scene and hasattr(context.scene.world, 'pco_general_settings'):
            world = context.scene.world
            if len(world.pco_planet_slots) > 0:
                planet_slots = world.pco_planet_slots
                planet_index = world.pco_active_planet_index
                planet_instance = planet_slots[planet_index]

                planet_data = planet_instance.planet_data

                layout = self.layout

                if world.pco_exposed.ring_instance < 0 or world.pco_exposed.ring_instance == planet_index:
                    layout.prop(planet_data, "rings_type", expand=True)
                    col = self.indentedColumn()
                    if planet_data.rings_type != 'DISABLED':
                        if planet_data.rings_type == 'TEXTURE':
                            col.template_ID(planet_data, "rings_texture", new="image.new", open="image.open")
                        elif planet_data.rings_type == 'PROCEDURAL':
                            col.prop(planet_data, "ring_procedural_texture_scale")
                            col.prop(planet_data, "ring_procedural_texture_detail")
                            col.prop(planet_data, "ring_procedural_texture_roughness")
                        col.prop(planet_data, "rings_tint")
                        col.prop(planet_data, "rings_vector", text="")
                        col.prop(planet_data, "rings_start")
                        col.prop(planet_data, "rings_end")
                        col.prop(planet_data, "rings_indirect_shadow")
                        col.prop(planet_data, "rings_translucency")
                        col.separator()
                elif world.pco_exposed.ring_instance < len(world.pco_planet_slots):
                    layout.scale_y = 0.6
                    layout.label(text='Rings are only supported', icon='ERROR')
                    warning_lines = [
                        "for one planet at a time! Currently",
                        "enabled on: " + planet_slots[world.pco_exposed.ring_instance].name,
                    ]
                    for line in warning_lines:
                        layout.label(text=line)


class PA_PT_PlanetRingsTB(PA_PCO_TB, Panel, PlanetRings):
    pass
